import numpy as np
from scipy.optimize import least_squares, fsolve
from qetpy.core._biasparams import get_biasparams_offsets, get_biasparams_ilg
from ._base_didv import _BaseDIDV, complexadmittance
from ._plot_didv import _PlotDIDV
from ._uncertainties_didv import get_smallsignalparams_vals, get_smallsignalparams_cov, get_smallsignalparams_sigmas, get_dPdI_with_uncertainties
from qetpy.utils import fft, ifft, fftfreq, rfftfreq
import copy
import warnings
warnings.simplefilter('default')


__all__ = [
    "didvinitfromdata",
    "DIDV",
]


def didvinitfromdata(tmean, didvmean, didvstd, offset, offset_err, fs, sgfreq,
                     sgamp, rsh, r0=0.3, rp=0.005, dutycycle=0.5,
                     add180phase=False, dt0=1.5e-6):
    """
    Function to initialize and process a dIdV dataset without having
    all of the traces, but just the parameters that are required for
    fitting. After running, this returns a DIDV class object that is
    ready for fitting.

    Parameters
    ----------
    tmean : ndarray
        The average trace in time domain, units of Amps
    didvstd : ndarray
        The complex standard deviation of the didv in frequency space
        for each frequency
    didvmean : ndarray
        The average trace converted to didv
    offset : float
        The offset (i.e. baseline value) of the didv trace, in Amps
    offset_err : float
        The error in the offset of the didv trace, in Amps
    fs : float
        Sample rate of the data taken, in Hz
    sgfreq : float
        Frequency of the signal generator, in Hz
    sgamp : float
        Peak-to-peak size of the square wave supplied by the signal
        generator, in Amps (equivalent to jitter in the QET bias)
    rsh : float
        Shunt resistance in the circuit, Ohms
    r0 : float, optional
        The estimated resistance of the TES in Ohms. Should be set if
        accurate small signal parameters are desired.
    rp : float, optional
        The estimated parasitic resistance of the non-shunt side of the
        TES circuit in Ohms. Should be set if accurate small signal
        parameters are desired.
    dutycycle : float, optional
        The duty cycle of the signal generator, should be a float
        between 0 and 1. Set to 0.5 by default
    add180phase : boolean, optional
        If the signal generator is out of phase (i.e. if it looks like
        --__ instead of __--), then this should be set to True. Adds
        half a period of the signal generator to the dt0 attribute.
    dt0 : float, optional
        The value of the starting guess for the time offset of the didv
        when fitting. The best way to use this value if it isn't
        converging well is to run the fit multiple times, setting `dt0`
        equal to the fit's next value, and seeing where the `dt0` value
        converges. The fit can have a difficult time finding the value
        on the first run if it the initial value is far from the actual
        value, so a solution is to do this iteratively.

    Returns
    -------
    didvobj : Object
        A DIDV class object that can be used to fit the dIdV and return
        the fit parameters.

    """

    didvobj = DIDV(
        None,
        fs,
        sgfreq,
        sgamp,
        rsh,
        r0=r0,
        rp=rp,
        add180phase=add180phase,
        dt0=dt0,
        dutycycle=dutycycle,
    )

    didvobj._didvmean = didvmean
    didvobj._didvstd = didvstd
    didvobj._offset = offset
    didvobj._offset_err = offset_err
    didvobj._tmean = tmean
    didvobj._dt0 = dt0

    if didvobj._add180phase:
        didvobj._dt0 = didvobj._dt0 + 1 / (2 * didvobj._sgfreq)

    didvobj._time = np.arange(len(tmean)) / fs - didvobj._dt0
    didvobj._freq = fftfreq(len(tmean), fs)
    
    nbins = len(didvobj._tmean)
    nperiods = np.floor(nbins*didvobj._sgfreq/didvobj._fs)

    flatindstemp = list()
    for i in range(0, int(nperiods)):
        # get index ranges for flat parts of trace
        flatindlow = int(
            (float(i) + 0.25) * didvobj._fs / didvobj._sgfreq
        ) + int(didvobj._dt0 * didvobj._fs)
        flatindhigh = int(
            (float(i) + 0.48) * didvobj._fs / didvobj._sgfreq
        ) + int(didvobj._dt0 * didvobj._fs)
        flatindstemp.append(range(flatindlow, flatindhigh))
    flatinds = np.array(flatindstemp).flatten()

    didvobj._flatinds = flatinds[np.logical_and(
        flatinds > 0, flatinds < nbins,
    )]

    return didvobj


class DIDV(_BaseDIDV, _PlotDIDV):
    """
    Class for fitting a didv curve for different types of models of the
    didv. Also gives various other useful values pertaining to the
    didv. This class supports doing 1, 2, and 3 pole fits. This is
    supported in a way that does one dataset at a time.

    """

    def __init__(self, rawtraces, fs, sgfreq, sgamp, rsh, tracegain=1.0,
                 r0=0.3, rp=0.005, dutycycle=0.5, add180phase=False,
                 dt0=None, autoresample=False):
        """
        Initialization of the DIDV class object

        Parameters
        ----------
        rawtraces : ndarray
            The array of rawtraces to use when fitting the didv. Should
            be of shape (number of traces, length of trace in bins).
            This can be any units, as long as tracegain will convert
            this to Amps.
        fs : float
            Sample rate of the data taken, in Hz
        sgfreq : float
            Frequency of the signal generator, in Hz
        sgamp : float
            Peak-to-peak size of the square wave supplied by the signal
            generator, in Amps (equivalent to jitter in the QET bias)
        rsh : float
            Shunt resistance in the circuit, Ohms
        tracegain : float, optional
            The factor that the rawtraces should be divided by to
            convert the units to Amps. If rawtraces already has units
            of Amps, then this should be set to 1.0
        r0 : float, optional
            The estimated resistance of the TES in Ohms. Should be set
            if accurate small signal parameters are desired.
        rp : float, optional
            The estimated parasitic resistance of the non-shunt side of
            the TES circuit in Ohms. Should be set if accurate small
            signal parameters are desired.
        dutycycle : float, optional
            The duty cycle of the signal generator, should be a float
            between 0 and 1. Set to 0.5 by default
        add180phase : boolean, optional
            If the signal generator is out of phase (i.e. if it looks
            like --__ instead of __--), then this should be set to
            True. Adds half a period of the signal generator to the
             `dt0` attribute
        dt0 : float, optional
            The value of the starting guess for the time offset of the
            didv when fitting. The best way to use this value if it
            isn't converging well is to run the fit multiple times,
            setting dt0 equal to the fit's next value, and seeing where
            the dt0 value converges. The fit can have a difficult time
            finding the value on the first run if it the initial value
            is far from the actual value, so a solution is to do this
            iteratively.
        autoresample : bool, optional
            If True, the initialization will automatically resample
            the data so that `fs` / `sgfreq` is an integer, which
            ensures that an arbitrary number of signal-generator
            periods can fit in an integer number of time bins. See
            `qetpy.utils.resample_data` for more info.

        """

        super().__init__(
            rawtraces,
            fs,
            sgfreq,
            sgamp,
            rsh,
            tracegain=tracegain,
            r0=r0,
            rp=rp,
            dutycycle=dutycycle,
            add180phase=add180phase,
            dt0=dt0,
            autoresample=autoresample,
        )


    @staticmethod
    def _fitdidv(freq, didv, yerr=None, A0=0.25, B0=-0.6, C0=-0.6,
                 tau10=-1.0/(2*np.pi*5e2), tau20=1.0/(2*np.pi*1e5), tau30=0.0,
                 dt=-10.0e-6, poles=2, isloopgainsub1=None,
                 bounds=None, lgcfix=None, verbose=0, max_nfev=1000,
                 method='trf', loss='linear',
                 ftol=1e-15, xtol=1e-15):
        """
        Function to find the fit parameters for either the 1-pole
        (A, tau2, dt), 2-pole (A, B, tau1, tau2, dt), or 3-pole
        (A, B, C, tau1, tau2, tau3, dt) fit.

        """

        if (poles==1):
            # assume the square wave is not inverted
            p0 = np.array((A0, tau20, dt), dtype=float)
            bounds1=bounds
            if bounds is None:
                bounds1 = (
                    np.array((0.0, 0.0, -np.inf)),
                    np.array((np.inf, np.inf, np.inf)),
                )
            # assume the square wave is inverted
            p02 =  np.array((-A0, tau20, dt), dtype=float)
            bounds2=bounds
            if bounds is None:
                bounds2 = (
                    np.array((-np.inf, 0.0, -np.inf)),
                    np.array((0.0, np.inf, np.inf)),
                )
        elif (poles==2):
            # assume loop gain > 1, where B<0 and tauI<0
            p0 = np.array((A0, B0, tau10, tau20, dt),
                          dtype=float)
            bounds1 = bounds
            if bounds is None:
                bounds1 = (
                    np.array((0.0, -np.inf, -np.inf, 0.0, -np.inf)),
                    np.array((np.inf, 0.0, 0.0, np.inf, np.inf)),
                )
                
            # assume loop gain < 1, where B>0 and tauI>0
            p02 = np.array((A0, -B0, -tau10, tau20, dt),
                           dtype=float)
            bounds2 = bounds
            if bounds is None:
                bounds2 = (
                    np.array((0.0, 0.0, 0.0, 0.0, -np.inf)),
                    np.array((np.inf, np.inf, np.inf, np.inf, np.inf)),
                )
            
            
        elif (poles==3):
            # assume loop gain > 1, where B<0 and tauI<0
            p0 = np.array((A0, B0, C0, tau10, tau20, tau30, dt),
                          dtype=float)
            bounds1 = bounds
            if bounds is None:
                bounds1 = (
                    np.array((0.0, -np.inf, -np.inf, -np.inf, 0.0, 0.0, -np.inf)),
                    np.array((np.inf, 0.0, 0.0, 0.0, np.inf, np.inf, np.inf)),
                )
            # assume loop gain < 1, where B>0 and tauI>0
            p02 = np.array((A0, -B0, -C0, -tau10, tau20, tau30, dt),
                           dtype=float)
            bounds2 = bounds
            if bounds is None:
                bounds2 = (
                    np.array((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -np.inf)),
                    np.array((np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)),
                )


        fix_params = None
        if lgcfix is not None:

            fix_params =  p0[lgcfix].copy()
            
            p0 = p0[~lgcfix]
            bounds1 = (bounds1[0][~lgcfix],
                       bounds1[1][~lgcfix])
            
            p02 = p02[~lgcfix]
            bounds2 = (bounds2[0][~lgcfix],
                       bounds2[1][~lgcfix])

            

            
        def _residual_calc(params):
            """
            Define a residual for the nonlinear least squares
            algorithm. Different functions for different amounts of
            poles.

            """
            if (poles==1):
                A, tau2, dt = params
                ci = DIDV._onepoleadmittance(
                    freq, A, tau2,
                ) * np.exp(-2.0j*np.pi*freq*dt)

            elif(poles==2):
                A, B, tau1, tau2, dt = params
                ci = DIDV._twopoleadmittance(
                    freq, A, B, tau1, tau2,
                ) * np.exp(-2.0j*np.pi*freq*dt)
                
            elif(poles==3):
                A, B, C, tau1, tau2, tau3, dt = params
                ci = DIDV._threepoleadmittance(
                    freq, A, B, C, tau1, tau2, tau3,
                ) * np.exp(-2.0j*np.pi*freq*dt)

            # the difference between the data and the fit
            diff = didv - ci
            # get the weights from yerr, these should be
            # 1/(standard deviation) for real and imaginary parts
            if (yerr is None):
                weights = 1.0+1.0j
            else:
                weights = 1.0/yerr.real+1.0j/yerr.imag
            # create the residual vector, splitting up real and imaginary
            # parts of the residual separately
            z1d = np.zeros(freq.size*2, dtype=np.float64)
            z1d[0:z1d.size:2] = diff.real*weights.real
            z1d[1:z1d.size:2] = diff.imag*weights.imag
            return z1d


        def _residual(var_params):
            """
            Function that is passed to nonlinear 
            least_squares scipy algorithm
            
            Parameters
            ----------

            var_params : np.array (dtype=float)
               variable fit parameters array

            
            Return
            ------

            residual :  np.array
            """

            if fix_params is None:
                return _residual_calc(var_params)
            else:
                all_params = np.zeros_like(lgcfix, dtype=float)
                np.place(all_params, lgcfix, fix_params)
                np.place(all_params, ~lgcfix, var_params)
                return _residual_calc(all_params)

        

        if (isloopgainsub1 is None):
            # res1 assumes loop gain > 1, where B<0 and tauI<0
            res1 = least_squares(
                _residual,
                p0,
                bounds=bounds1,
                loss=loss,
                max_nfev=max_nfev,
                x_scale=np.abs(p0),
                verbose=verbose,
                method=method,
                ftol=ftol,
                xtol=xtol,
            )
            # res2 assumes loop gain < 1, where B>0 and tauI>0
            res2 = least_squares(
                _residual,
                p02,
                bounds=bounds2,
                loss=loss,
                max_nfev=max_nfev,
                x_scale=np.abs(p0),
                verbose=verbose,
                method=method,
                ftol=ftol,
                xtol=xtol,
            )
            # check which loop gain cases gave the better fit
            if (res1['cost'] < res2['cost']):
                res = res1
            else:
                res = res2
        elif isloopgainsub1:
            # assume loop gain < 1, where B>0 and tauI>0
            res = least_squares(
                _residual,
                p02,
                bounds=bounds2,
                loss=loss,
                max_nfev=max_nfev,
                x_scale=np.abs(p0),
                verbose=verbose,
                method=method,
                ftol=ftol,
                xtol=xtol,
            )
        else:
            #assume loop gain > 1, where B<0 and tauI<0
            res = least_squares(
                _residual,
                p0,
                bounds=bounds1,
                loss=loss,
                max_nfev=max_nfev,
                x_scale=np.abs(p0),
                verbose=verbose,
                method=method,
                ftol=ftol,
                xtol=xtol,
            )


        # variables
        popt = res['x'].copy()
        if lgcfix is not None:
            popt = np.zeros_like(lgcfix, dtype=float)
            np.place(popt, lgcfix, fix_params)
            np.place(popt, ~lgcfix, res['x'])
            
        # cost (chi2 = 2*cost)
        cost = 2*res['cost']
        fun = res['fun']
    
        # check if the fit failed (usually only happens when we reach maximum
        # evaluations, likely when fitting assuming the wrong loop gain)
        if not res['success'] :
            print(f"{poles}-Pole Fit Failed: " + res['message'])

        # take matrix product of transpose of jac and jac, take the inverse
        # to get the analytic covariance matrix
        pcovinv = np.dot(res["jac"].transpose(), res["jac"])
        pcov = np.linalg.pinv(pcovinv)

        return popt, pcov, cost


    def dofit(self, poles, fcutoff=np.inf,
              bounds=None, guess_params=None,
              guess_isloopgainsub1=None,
              lgcfix=None, verbose=0, max_nfev=1000,
              method='trf', loss='linear',
              ftol=1e-15, xtol=1e-15):
        """
        This method does the fit that is specified by the variable
        poles. If the `processtraces` method has not been run yet, then
        this method will run that first.

        Parameters
        ----------
        poles : int
            The fit that should be run. Should be 1, 2, or 3.
        fcutoff : float, optional
            The cutoff frequency in Hz, above which data is ignored in
            the specified fitting routine. Default is `np.inf`, which
            is equivalent to no cutoff frequency.
        
        FIXME: Missing parameters here

        Raises
        ------
        ValueError
            If the inputted `poles` is not 1, 2, or 3.

        Notes
        -----
        Depending on the fit, there are three possible models to be
        used with different parameterizations:

        1-pole model
            - has the form:
                dV/dI = A * (1.0 + 2.0j * pi * freq * tau2)

        2-pole model
            - has the form:
                dV/dI = A * (1.0 + 2.0j * pi * freq * tau2)
                      + B / (1.0 + 2.0j * pi * freq * tau1)

        3-pole model
            - note the placement of the parentheses in the last term of
              this model, such that pole related to `C` is in the
              denominator of the `B` term
            - has the form: 
                dV/dI = A * (1.0 + 2.0j * pi * freq * tau2)
                      + B / (1.0 + 2.0j * pi * freq * tau1
                      - C / (1.0 + 2.0j * pi * freq * tau3))

        """

        if self._tmean is None:
            self.processtraces()

        fit_freqs = np.abs(self._freq) < fcutoff
             
        # 1-Pole fit
        if poles==1:
            
            # guess the 1 pole square wave parameters
            A0_1pole, tau20_1pole = DIDV._squarewaveguessparams(
                self._tmean,
                self._sgamp,
                self._rsh,
            )
            
            # time shift guess
            dt = self._dt0
      
            # overrite guessed values if provided by user
            if guess_params is not None:
                if len(guess_params) != 3:
                    raise ValueError(
                        'Expecting 2 guessed parameters. '
                        + 'Found ' + str(len(guess_params)))
                inA0, intau20, indt = guess_params
                if inA0 is not None:
                    A0_1pole = inA0
                if intau20 is not None:
                    tau20_1pole = intau20
                if indt is not None:
                    dt = indt

            
            # 1 pole fitting
            fitparams1, fitcov1, fitcost1 = DIDV._fitdidv(
                self._freq[fit_freqs],
                self._didvmean[fit_freqs],
                yerr=self._didvstd[fit_freqs],
                A0=A0_1pole,
                tau20=tau20_1pole,
                dt=dt,
                poles=poles,
                isloopgainsub1=False,
                bounds=bounds,
                lgcfix=lgcfix,
                verbose=verbose,
                max_nfev=max_nfev,
                method=method,
                loss=loss,
                ftol=ftol,
                xtol=xtol,
            )

            # Convert to didv falltimes
            falltimes1 = DIDV._findpolefalltimes(fitparams1)

            # cost: divide by NDOF
            fitcost1 /= (np.sum(fit_freqs)-len(fitparams1))
            
            # store as dictionary
            self._fit_results[1] = DIDV._fitresult(
                poles,
                fitparams1,
                fitcov1,
                falltimes1,
                fitcost1,
                lgcfix=lgcfix,
            )

            # store offset 
            self._fit_results[1]['offset'] = self._offset
            self._fit_results[1]['offset_err'] = self._offset_err

        elif poles==2:
  
            
            # Guess the starting parameters for 2 pole fitting
            A0, B0, tau10, tau20, isloopgainsub1 = DIDV._guessdidvparams(
                self._tmean,
                self._tmean[self._flatinds],
                self._sgamp,
                self._rsh,
                L0=1.0e-7,
            )

            # time shift
            dt0 = self._dt0
            
            # overrite guessed values if provided by user
            if guess_params is not None:
                if len(guess_params) != 5:
                    raise ValueError(
                        'Expecting 5 guessed parameters. '
                        + 'Found ' + str(len(guess_params)))
                inA0, inB0, intau10, intau20, indt = guess_params
                if inA0 is not None:
                    A0 = inA0
                if inB0 is not None:
                    B0 = inB0
                if intau10 is not None:
                    tau10 = intau10
                if intau20 is not None:
                    tau20 = intau20
                if indt is not None:
                    dt0 = indt

            # loopgainsub1   
            if guess_isloopgainsub1 is not None:
                isloopgainsub1 = guess_isloopgainsub1
       
            # 2 pole fitting
            fitparams2, fitcov2, fitcost2 = DIDV._fitdidv(
                self._freq[fit_freqs],
                self._didvmean[fit_freqs],
                yerr=self._didvstd[fit_freqs],
                A0=A0,
                B0=B0,
                tau10=tau10,
                tau20=tau20,
                dt=dt0,
                poles=poles,
                isloopgainsub1=isloopgainsub1,
                bounds=bounds,
                lgcfix=lgcfix,
                verbose=verbose,
                max_nfev=max_nfev,
                method=method,
                loss=loss,
                ftol=ftol,
                xtol=xtol,
            )

            # full coveriance matrix
            full_fitcov2 = fitcov2
            if lgcfix is not None:
                full_fitcov2 = np.zeros((5, 5))
                free_indices = np.where(~lgcfix)[0]
                full_fitcov2[np.ix_(free_indices, free_indices)] = fitcov2
            
            # cost: divide by NDOF
            fitcost2 = fitcost2/(np.sum(fit_freqs)-len(fitparams2))

            # Convert to didv falltimes
            falltimes2 = DIDV._findpolefalltimes(fitparams2)
            
            # Convert to a 3-pole model with  C=0  and tau3=0

            # 1)  fit parameters
            indices = [2, 4]
            fitparams2 = np.insert(fitparams2, [2,4], 0)
            
            if lgcfix is not None:
                if isinstance(lgcfix, list):
                    lgcfix.insert(2, True)
                    lgcfix.insert(5, True)
                elif isinstance(lgcfix, np.ndarray):
                    lgcfix = np.insert(lgcfix, [2,4], 0)
                    
            # 2) cov
                
            # Insert a row of zeros at index 2
            intermediate_cov2 = np.insert(full_fitcov2, 2, 0, axis=0)
            # Insert a column of zeros at index 2
            intermediate_cov2 = np.insert(intermediate_cov2, 2, 0, axis=1) 

            # For tau3, insert zeros at index 5
            # Now index 5 for rows
            final_cov2 = np.insert(intermediate_cov2, 5, 0, axis=0)
            # And index 5 for columns
            final_cov2 = np.insert(final_cov2, 5, 0, axis=1) 

            # store as dictionary
            self._fit_results[2] = DIDV._fitresult(
                poles,
                fitparams2,
                final_cov2,
                falltimes2,
                fitcost2,
                lgcfix=lgcfix)

            # store a few more parameters
            self._fit_results[2]['offset'] = self._offset
            self._fit_results[2]['offset_err'] = self._offset_err

        elif poles==3:
            
            if (self._fit_results[2] is None
                or 'params' not in self._fit_results[2]):
                
                # Guess the 3-pole fit
                A0, B0, tau10, tau20 = DIDV._guessdidvparams(
                    self._tmean,
                    self._tmean[self._flatinds],
                    self._sgamp,
                    self._rsh,
                    L0=1.0e-7,
                )[:-1]
                B0 = -abs(B0)
                C0 = -0.05
                tau10 = -abs(tau10)
                tau30 = 1.0e-3
                dt0 = self._dt0
                
            else:
                A0 = self._fit_results[2]['params']['A']
                B0 = -abs(self._fit_results[2]['params']['B'])
                C0 = -0.05
                tau10 = -abs(self._fit_results[2]['params']['tau1']) 
                tau20 = self._fit_results[2]['params']['tau2']
                tau30 = 1.0e-3
                dt0 = self._fit_results[2]['params']['dt']
                
            # is loop gain < 1
            isloopgainsub1 = DIDV._guessdidvparams(
                self._tmean,
                self._tmean[self._flatinds],
                self._sgamp,
                self._rsh,
                L0=1.0e-7,
            )[-1]

        
            # overwrite guessed values if provided by user
            if guess_params is not None:
                if len(guess_params) != 7:
                    raise ValueError(
                        'Expecting 7 guessed parameters. '
                        + 'Found ' + str(len(guess_params)))
                inA0, inB0, inC0, intau10, intau20, intau30, indt = guess_params
                if inA0 is not None:
                    A0 = inA0
                if inB0 is not None:
                    B0 = inB0
                if inC0 is not None:
                    C0 = inC0
                if intau10 is not None:
                    tau10 = intau10
                if intau20 is not None:
                    tau20 = intau20
                if intau30 is not None:
                    tau30 = intau30 
                if indt is not None:
                    dt0 = indt

            # loopgainsub1   
            if guess_isloopgainsub1 is not None:
                isloopgainsub1 = guess_isloopgainsub1
            
            # 3 pole fitting

            fitparams3, fitcov3, fitcost3 = DIDV._fitdidv(
                self._freq[fit_freqs],
                self._didvmean[fit_freqs],
                yerr=self._didvstd[fit_freqs],
                A0=A0,
                B0=B0,
                C0=C0,
                tau10=tau10,
                tau20=tau20,
                tau30=tau30,
                dt=dt0,
                poles=poles,
                isloopgainsub1=isloopgainsub1,
                bounds=bounds,
                lgcfix=lgcfix,
                verbose=verbose,
                max_nfev=max_nfev,
                method=method,
                loss=loss,
                ftol=ftol,
                xtol=xtol,
            )

            # cost: divide by NDOF
            fitcost3 = fitcost3/(np.sum(fit_freqs)-len(fitparams3))

            # full coveriance matrix
            full_fitcov3 = fitcov3
            if lgcfix is not None:
                full_fitcov3 = np.zeros((7, 7))
                free_indices = np.where(~lgcfix)[0]
                full_fitcov3[np.ix_(free_indices, free_indices)] = fitcov3
            
            # Convert to didv falltimes
            falltimes3 = DIDV._findpolefalltimes(fitparams3)

            # store results
            self._fit_results[3] = DIDV._fitresult(
                poles,
                fitparams3,
                full_fitcov3,
                falltimes3,
                fitcost3,
                lgcfix=lgcfix,
            )

            # store offset 
            self._fit_results[3]['offset'] = self._offset
            self._fit_results[3]['offset_err'] = self._offset_err
     
        else:
            raise ValueError("The number of poles should be 1, 2, or 3.")

    def calc_smallsignal_params(self, biasparams=None,
                                poles=None,
                                lgc_verbose=True,
                                lgc_diagnostics=False):
        """
        Calculate small signal parametres and their uncertainties 
        using prior biasparams (I0, R0) calculatiobns.
        
        FIXME: add parameters description!
        """

        # check which models have been fitted (1, 2 and/or 3 poles)
        model_list  = list()

        if poles is not None:

            if isinstance(poles, list):
                model_list = poles
            else:
                model_list = [poles]

            for model_poles in model_list:
                if (self._fit_results[model_poles] is None
                    or 'params' not in  self._fit_results[model_poles]):
                    raise ValueError(f'ERROR: {model_poles}-poles fit needs to '
                                     f'be done first!')
        else:

            if (self._fit_results[1] is not None
                and 'params' in  self._fit_results[1]):
                model_list.append(1)
                
            if (self._fit_results[2] is not None
                and 'params' in  self._fit_results[2]):
                model_list.append(2)

            if (self._fit_results[3] is not None
                and 'params' in  self._fit_results[3]):
                model_list.append(3)
                
                
        if not model_list:
            print('WARNING: No fit have been done. Doing nothing...')
            return

        # 1-pole (does not require biasparams)
        if 1 in model_list:

            print(f'INFO: Calculating small signal parameters '
                  f'for 1-pole model!')
            
            # calc
            self._calc_ssp(1, biasparams_dict=None,
                           lgc_ssp_light=False)

            # remove from list
            model_list.remove(1)

            if not model_list:
                return

        # 2 and/or 3 poles
            
        # check ivsweep results and convert to dictionary
        if (biasparams is None or not isinstance(biasparams, dict)):
            raise ValueError(f'ERROR: "biasparams" is required '
                             f'and should be a dictionary!')
        
        required_parameters = ['rp', 'i0', 'i0_err','r0', 'r0_err']
        for par in required_parameters:
            if par not in biasparams.keys():
                raise ValueError(f'ERROR: parameter {par} not found in '
                                 '"ivsweep_results" dictionary!')

        # initialize  bias parameters dict
        biasparams_dict = biasparams.copy()
        self._r0 = biasparams_dict['r0']
        self._rp = biasparams_dict['rp']


        # calculate small signal parameters
        
        # loop poles
        for model_poles in model_list:

            print(f'INFO: Calculating small signal parameters '
                  f'for {model_poles}-poles model ')
            
            # calc
            self._calc_ssp(model_poles,
                           biasparams_dict=biasparams_dict.copy(),
                           lgc_ssp_light=True)
            
 
    @staticmethod
    def _fitresult(poles, params_array, cov, falltimes, cost,                   
                   lgcfix=None):

        """
        Function for converting data from different fit results to a
        results dictionary.

        """

        result = dict()
        result['lgcfix'] = lgcfix
        result['params_array'] = params_array
        
        # errors
        errors = np.diag(cov)**0.5
        if lgcfix is not None:
            errors = np.zeros_like(lgcfix, dtype=float)
            np.place(errors, lgcfix, 0.0)
            np.place(errors, ~lgcfix,  np.diag(cov)**0.5)
        
        if poles == 1:
            result['params'] = {
                'A': params_array[0],
                'tau2': params_array[1],
                'dt': params_array[2],
                'B':0,
                'C':0,
                'tau1':0,
                'tau3':0
            }
            result['cov'] = cov
            result['errors'] = {
                'A': errors[0],
                'tau2': errors[1],
                'dt': errors[2],
                'B':0,
                'C':0,
                'tau1':0,
                'tau3':0
            }

        elif (poles == 2 or poles == 3):

            result['params'] = {
                'A': params_array[0],
                'B': params_array[1],
                'C': params_array[2],
                'tau1': params_array[3],
                'tau2': params_array[4],
                'tau3': params_array[5],
                'dt': params_array[6],
            }
            
            result['cov'] = cov
            
            result['errors'] = {
                'A': errors[0],
                'B': errors[1],
                'C': errors[2],
                'tau1': errors[3],
                'tau2': errors[4],
                'tau3': errors[5],
                'dt': errors[6],
            }

        # other params
        result['falltimes'] = falltimes
        result['cost'] = cost
      
        return result

    def _calc_ssp(self, poles,
                  biasparams_dict=None,
                  lgc_ssp_light=False):
        """
        Function to calculate small signal parameters  from fit result
        """
        
        # check if r0/rp available in biasparams_dict
        if biasparams_dict is not None:

            # copy
            biasparams_dict = copy.deepcopy(biasparams_dict)
                    
            if 'r0' in biasparams_dict:
                self._r0 = biasparams_dict['r0']
            if 'rp' in biasparams_dict:
                self._rp = biasparams_dict['rp']
                
        # 1-poles fit
        if poles == 1:
            if (self._fit_results[1] is None
                or 'params' not in self._fit_results[1]):
                raise ValueError(
                    'ERROR: No 1-poles fit done! Unable to '
                    'calculate small signal parameters...')
                    
            smallsignalparams = DIDV._converttotesvalues(
                self._fit_results[1]['params'],
                rsh=self._rsh, r0=self._r0, rp=self._rp
            )

            self._fit_results[1]['smallsignalparams'] = (
                smallsignalparams.copy()
            )
            
            self._fit_results[1]['didv0'] = (
                complexadmittance(0, self._fit_results[1]['smallsignalparams']).real
            )
            
            # store bias params
            self._fit_results[1]['biasparams'] = biasparams_dict
            
                
        # 2-poles and 3-poles fit
        if poles == 2 or poles == 3:
            # get result dictionary
            if (self._fit_results[poles] is None
                or 'params' not in self._fit_results[poles]):
                raise ValueError(
                    f'ERROR: No {poles}-poles fit done! Unable to '
                    f'calculate small signal parameters...')
                    
            # results
            results = copy.deepcopy(self._fit_results[poles])
            
            # store bias params
            results['biasparams'] = biasparams_dict
                      
            # convert fit parameterts to smallsignalparams
            smallsignalparams = DIDV._converttotesvalues(
                results['params'],
                rsh=self._rsh, r0=self._r0, rp=self._rp
            )
            
            results['smallsignalparams'] = smallsignalparams.copy()
                
            # calculate small signal parameters cov/sigmas
            if lgc_ssp_light:
                
                if biasparams_dict is None:
                    raise ValueError(
                        'ERROR: "biasparams_dict" required when '
                        'lgc_ssp_light=True'
                    )
                    
                ssp_light_vals = get_smallsignalparams_vals(results)
                ssp_light_cov = get_smallsignalparams_cov(results)
                ssp_light_sigmas = get_smallsignalparams_sigmas(results)
                
                results['ssp_light'] = {
                    'vals': ssp_light_vals,
                    'cov': ssp_light_cov,
                    'sigmas': ssp_light_sigmas,
                }
                
            results['didv0'] = (
                complexadmittance(0, results['smallsignalparams']).real
            )
            
            dpdi, dpdi_err = get_dPdI_with_uncertainties([0.0], results)
            results['dpdi0'] = (dpdi[0].real)
            results['dpdi0_err'] = (dpdi_err[0].real)
                
            # store results internally
            self._fit_results[poles] = results



import pytest
import numpy as np

from helpers import isclose, create_example_data, create_example_muontail
import qetpy as qp
from qetpy.core._fitting import _argmin_chi2, _get_pulse_direction_constraint_mask

def test_interpolation():
    """
    Testing function for the precomputed parabolic interpolation in the
    `OptimumFilter` class.

    """

    vals = [2, 0, 1]
    bestind = 1
    delta = 1

    output1 = qp.OptimumFilter._interpolate_parabola(vals, bestind, delta)

    assert np.all(np.asarray(output1)==np.array([1 / 6, -1 / 24]))

    output2 = qp.OptimumFilter._interpolate_parabola(
        vals, bestind, delta, t_interp=output1[0],
    )

    assert np.all(np.asarray(output1)==np.array([1 / 6, -1 / 24]))

    output3 = qp.OptimumFilter._interpolate_of(vals, vals, bestind, delta)

    output_comp = np.array([output2[1], output1[0], output1[1]])
    assert np.all(output_comp==np.asarray(output3))

def test_get_pulse_direction_constraint_mask():
    """
    Testing function for `qetpy.core._fitting._get_pulse_direction_constraint_mask`.
    
    """
    
    amps = np.array([1, 0.1, -0.1, 3])
    
    out1 = _get_pulse_direction_constraint_mask(amps, pulse_direction_constraint=0)
    assert out1 is None
    
    out2 = _get_pulse_direction_constraint_mask(amps, pulse_direction_constraint=1)
    assert np.all(out2 == np.array([1, 1, 0, 1], dtype=bool))
    
    out3 = _get_pulse_direction_constraint_mask(amps, pulse_direction_constraint=-1)
    assert np.all(out3 == np.array([0, 0, 1, 0], dtype=bool))
    
    with pytest.raises(ValueError):
        out4 = _get_pulse_direction_constraint_mask(amps, pulse_direction_constraint=2)
        

def test_argmin_chi2():
    """
    Testing function for `qetpy.core._fitting._argmin_chi2`.
    
    """
    
    x = np.array([-0.1, -1, 0.1, 1])
    
    res1 = _argmin_chi2(x)
    assert res1 == 1
    res2 = _argmin_chi2(x, nconstrain=2)
    assert res2 == 1
    res3 = _argmin_chi2(x, nconstrain=2, lgcoutsidewindow=True)
    assert res3 == 0
    res4 = _argmin_chi2(x, nconstrain=1000)
    assert res4 == 1
    res5 = _argmin_chi2(x, constraint_mask=np.array([False, False, False, False]))
    assert np.isnan(res5)
    
    with pytest.raises(ValueError):
        res6 = _argmin_chi2(x, nconstrain=-1)
    
    amps = np.array([1, 0.1, -0.1, 3])
    
    constraint_mask = _get_pulse_direction_constraint_mask(amps, pulse_direction_constraint=1)
    
    res1 = _argmin_chi2(x, constraint_mask=constraint_mask)
    assert res1 == 1
    res2 = _argmin_chi2(x, nconstrain=2, constraint_mask=constraint_mask)
    assert res2 == 1
    res3 = _argmin_chi2(x, nconstrain=2, lgcoutsidewindow=True, constraint_mask=constraint_mask)
    assert res3 == 0
    res4 = _argmin_chi2(x, nconstrain=2, lgcoutsidewindow=True, constraint_mask=np.array([False, False, False, False]))
    assert np.isnan(res4)
        
    res1 = _argmin_chi2(x, nconstrain=2, windowcenter=1)
    assert res1 == 2
    res2 = _argmin_chi2(x, nconstrain=2, lgcoutsidewindow=True,  windowcenter=1)
    assert res2 == 1
    
    with pytest.raises(ValueError):
        res3 = _argmin_chi2(x, nconstrain=1, windowcenter=4)
    
    
def test_OptimumFilter():
    """
    Testing function for `qetpy.OptimumFilter` class.
    
    """

    signal, template, psd = create_example_data()
    fs = 625e3

    OF = qp.OptimumFilter(signal, template, psd, fs)
    res = OF.ofamp_nodelay()
    #assert isclose(res, (-1.589803642041125e-07, 2871569.457990007), rtol=1e-6)
    
    res = OF.energy_resolution()
    #assert isclose(res, 2.3725914280425287e-09, rtol=1e-6)

    res = OF.ofamp_withdelay()
    #assert isclose(res, (4.000884927004103e-06, 0.00016, 32474.45440205792), rtol=1e-6)
    
    res2 = OF.ofamp_nodelay(windowcenter=int(res[1] * fs))
    #assert isclose(res2, res[::2], rtol=1e-6)

    res = OF.ofamp_withdelay(interpolate_t0=True)
    #assert isclose(res, (4.000884983533612e-06, 0.00016000193750599913, 32474.375405751092), rtol=1e-6)

    res = OF.time_resolution(res[0])
    #assert isclose(res, 5.746611055379949e-09, rtol=1e-6)
    
    res = OF.ofamp_withdelay(nconstrain=100)
    #assert isclose(res, (6.382904231454342e-07, 7.84e-05, 2803684.0424425197), rtol=1e-6)

    res = OF.ofamp_withdelay(nconstrain=100, lgcoutsidewindow=True)
    #assert isclose(res, (4.000884927004103e-06, 0.00016, 32474.45440205792), rtol=1e-6)
    
    res = OF.ofamp_withdelay(nconstrain=3 ,windowcenter=-5)
    #assert isclose(res, (-1.748136068514983e-07, -9.6e-06, 2870630.5945196496), rtol=1e-6)
    
    res = OF.chi2_lowfreq(amp=4.000884927004103e-06, t0=0.00016, fcutoff=10000)
    #assert isclose(res, 1052.9089578293142, rtol=1e-6)
    
    res = OF.chi2_nopulse()
    #assert isclose(res, 2876059.4034037213, rtol=1e-6)
    
    OF.update_signal(signal)
    res = OF.ofamp_pileup_stationary()
    #assert isclose(res, (2.884298804131357e-09, 4.001001614298674e-06, 0.00016, 32472.978956471197), rtol=1e-6)
    
    signal, template, psd = create_example_data(lgcpileup=True)
    
    OF.update_signal(signal)
    res1 = OF.ofamp_withdelay()
    res = OF.ofamp_pileup_iterative(res1[0], res1[1])
    #assert isclose(res, (4.000882414471985e-06, 0.00016, 32477.55571848713), rtol=1e-6)

    res = OF.ofamp_pileup_iterative(res1[0], res1[1], interpolate_t0=True)
    #assert isclose(res, (4.000882471002715e-06, 0.00016000193750598994, 32477.47671986579), rtol=1e-6)

    res = OF.ofamp_pileup_iterative(res1[0], res1[1], nconstrain=100, lgcoutsidewindow=False)
    #assert isclose(res, (6.382879106117655e-07, 7.84e-05, 2803684.142039136), rtol=1e-6)

    res = OF.ofamp_pileup_iterative(res1[0], res1[1], nconstrain=100, lgcoutsidewindow=True)
    #assert isclose(res, (4.000882414471985e-06, 0.00016, 32477.55571848713), rtol=1e-6)

    signal, template, psd = create_example_data(lgcbaseline=True)

    OF.update_signal(signal)
    res = OF.ofamp_baseline()
    #assert isclose(res, (4.000884927004102e-06, 0.00016, 32474.454402058076), rtol=1e-6)

    res = OF.ofamp_baseline(interpolate_t0=True)
    #assert isclose(res, (4.000884983446686e-06, 0.00016000193752042192, 32474.37540399214), rtol=1e-6)

    res = OF.ofamp_baseline(nconstrain=100)
    #assert isclose(res, (6.434754982839688e-07, 7.84e-05, 2806781.3450564747), rtol=1e-6)

    res = OF.ofamp_baseline(nconstrain=100, lgcoutsidewindow=True)
    #assert isclose(res, (4.000884927004102e-06, 0.00016, 32474.454402058076), rtol=1e-6)


def test_ofamp():
    """
    Testing function for `qetpy.ofamp`.
    
    """
    
    signal, template, psd = create_example_data()
    fs = 625e3
    
    res1 = qp.ofamp(signal, template, psd, fs, lgcsigma=True, nconstrain=100, withdelay=True)
    
    OF = qp.OptimumFilter(signal, template, psd, fs)
    res2 = OF.ofamp_withdelay(nconstrain=100)
    
    res_compare1 = res2 + (OF.energy_resolution(), )
    
    res3 = qp.ofamp(signal, template, psd, fs, withdelay=False)
    
    OF = qp.OptimumFilter(signal, template, psd, fs)
    res4 = OF.ofamp_nodelay()
    
    res_compare2 = (res4[0], 0.0, res4[1])
    
    assert isclose(res1, res_compare1)
    assert isclose(res3, res_compare2)
    
def test_ofamp_pileup():
    """
    Testing function for `qetpy.ofamp_pileup`.
    
    """
    signal, template, psd = create_example_data(lgcpileup=True)
    fs = 625e3
    
    res1 = qp.ofamp_pileup(signal, template, psd, fs)

    OF = qp.OptimumFilter(signal, template, psd, fs)
    res2 = OF.ofamp_withdelay()
    res3 = OF.ofamp_pileup_iterative(res2[0], res2[1])

    res_compare = res2[:-1] + res3
    
    assert isclose(res1, res_compare)
    
def test_ofamp_pileup_stationary():
    """
    Testing function for `qetpy.ofamp_pileup_stationary`.
    
    """
    
    signal, template, psd = create_example_data()
    fs = 625e3
    
    res1 = qp.ofamp_pileup_stationary(signal, template, psd, fs)

    OF = qp.OptimumFilter(signal, template, psd, fs)
    res2 = OF.ofamp_pileup_stationary()
    
    assert isclose(res1, res2)
    
def test_chi2lowfreq():
    """
    Testing function for `qetpy.chi2lowfreq`.
    
    """
    
    signal, template, psd = create_example_data()
    fs=625e3
    
    res1 = qp.ofamp(signal, template, psd, fs)

    chi2low = qp.chi2lowfreq(signal, template, res1[0], res1[1], psd, fs, fcutoff=10000)
    
    OF = qp.OptimumFilter(signal, template, psd, fs)
    res2 = OF.ofamp_withdelay()
    chi2low_compare = OF.chi2_lowfreq(amp=res2[0], t0=res2[1], fcutoff=10000)
    
    assert isclose(chi2low, chi2low_compare)
    
def test_chi2_nopulse():
    """
    Testing function for `qetpy.chi2_nopulse`.
    
    """
    
    signal, template, psd = create_example_data()
    fs=625e3
    
    res1 = qp.chi2_nopulse(signal, psd, fs)
        
    OF = qp.OptimumFilter(signal, template, psd, fs)
    res2 = OF.chi2_nopulse()
    
    assert isclose(res1, res2)
    
def test_OFnonlin():
    """
    Testing function for `qetpy.OFnonlin` class.
    
    """

    signal, template, psd = create_example_data()
    fs=625e3
    
    signal = np.roll(signal, -100) # undo the roll in create_example_data to make test easier
    
    nlin = qp.OFnonlin(psd, fs, template=template)
    res1a = nlin.dofit(signal, npolefit=1, lgcfullrtn=False, lgcplot=True, taurise=20e-6, scale_amplitude=True)
    res1 = nlin.dofit(signal, npolefit=1, lgcfullrtn=False, lgcplot=True, taurise=20e-6, scale_amplitude=False)
    res2a = nlin.dofit(signal, npolefit=2, lgcfullrtn=False, lgcplot=True, scale_amplitude=True)
    res2 = nlin.dofit(signal, npolefit=2, lgcfullrtn=False, lgcplot=True, scale_amplitude=False)
    res3 = nlin.dofit(signal, npolefit=3, lgcfullrtn=False, lgcplot=True)
    res4 = nlin.dofit(signal, npolefit=4, lgcfullrtn=False, lgcplot=True)


    """
    assert isclose(res1a, [4.00801301e-06, 6.57824204e-05, 2.60000272e-02], rtol=1e-6)
    assert isclose(res1, [9.69052142e-06, 6.57726246e-05, 2.60000311e-02], rtol=1e-6)
    assert isclose(res2a, [4.00954923e-06, 1.98364701e-05,
                           6.60692408e-05, 2.60000622e-02], rtol=1e-6)
    assert isclose(res2, [9.35735641e-06, 1.93841450e-05,
                          6.69374757e-05, 2.60001467e-02], rtol=1e-6)
    assert isclose(res3, [9.63212105e-06, 7.84008525e-08,
                          1.99470233e-05, 6.51290402e-05,
                          1.49376493e-04, 2.60000597e-02], rtol=1e-6)
    assert isclose(res4, [9.15466859e-06, 5.06330471e-07,
                          1.68478560e-08, 1.97920064e-05,
                          6.38075778e-05, 1.01482898e-04,
                          1.92735955e-04, 2.60001024e-02], rtol=1e-6)
    """

    
def test_MuonTailFit():
    """
    Testing function for `qetpy.MuonTailFit` class.
    
    """
    
    signal, psd = create_example_muontail()    
    fs = 625e3
    
    mtail = qp.MuonTailFit(psd, fs)
    res = mtail.fitmuontail(signal, lgcfullrtn=False)
    """
    assert isclose(
        res,
        [4.5964397140760587e-07, 0.0201484585061281],
        rtol=1e-8,
    )
    """



import warnings
import itertools
from packaging import version
import numpy as np
from scipy import optimize
from scipy import __version__ as SCIPY_VERSION

import qetpy as qp


__all__ = [
    "PileupDE",
]


class PileupDE(object):
    """
    Class for determining the amplitudes and times of an arbitrary
    number of pulses via the Optimum Filter (OF) formalism.

    Attributes
    ----------
    pileup_res : ndarray, NoneType
        Calculated by the `run` class method, this contains the
        amplitudes, time-shift values, and chi-square from the
        pileup optimum filter within the range of time values
        where the pulses would have a large amount of interference,
        nonnegligibly reducing the sensitivity of the iterative pileup
        optimum filter. Set to None when initialized or when the signal
        is updated via the `update_signal` class method.

    Notes
    -----
    The parameters in `pileup_res` are:

        (amplitude 1, time offset 1,
         amplitude 2, time offset 2, 
         ...,
         chi-square)

    """

    def __init__(self, signal, template, psd, fs, error_cutoff=0.1,
                 ac_coupled=True, integralnorm=False):
        """
        Initalization of the PileupOF class.

        Parameters
        ----------
        signal : ndarray
            The signal that we want to apply the pileup optimum filter
            to (units should be Amps).
        template : ndarray
            The pulse template to be used for the pileup optimum filter
            (should be normalized to a max height of 1 beforehand).
        psd : ndarray
            The two-sided psd that will be used to describe the noise
            in the signal (in Amps^2/Hz)
        fs : ndarray
            The sample rate of the data being taken (in Hz).
        error_cutoff : float
            The cutoff on the error in the inversion of the pileup OF
            matrix to determine the time range to consider pileup.
            Setting to small values corresponds to a larger time
            range. Default is 0.1.
        ac_coupled : bool, optional
            If True, the zero frequency bin of the psd
            will be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If False, then the zero frequency
            bin is kept. Default is True.
        integralnorm : bool, optional
            If set to True, then the template will be normalized to
            have an integral of 1, and any optimum filters will instead
            return the optimum integral in units of Coulombs. If set to
            False, then the usual optimum filter amplitudes will be
            returned (in units of Amps). Default is False.

        """

        self._psd = np.zeros(len(psd))
        self._psd[:] = psd
        self._psd0 = psd[0]

        if ac_coupled:
            self._psd[0] = np.inf

        self._nbins = signal.shape[-1]
        self._fs = fs
        self._df = self._fs / self._nbins

        self._s = np.fft.fft(template) / self._nbins / self._df

        if integralnorm:
            self._s /= self._s[0]

        self._phi = self._s.conjugate() / self._psd
        self._norm = np.real(np.dot(self._phi, self._s)) * self._df

        self._time_array = None
        self._pmatrix_off = np.real(
            np.fft.ifft(self._s * self._phi) / self._norm * self._fs
        )[:self._nbins//2]

        self._freqs = np.fft.fftfreq(self._nbins, d=1 / self._fs)

        self._tcutoff = self._determine_tcutoff(error_cutoff)

        self._OF = qp.OptimumFilter(
            signal,
            template,
            psd,
            fs,
            coupling="AC" if ac_coupled else "DC",
            integralnorm=integralnorm,
        )

        self._update_signal(signal)


    def _update_signal(self, signal):
        """
        Hidden function for recalculating parameters that depend on the
        signal.

        """

        self.pileup_res = None
        self._v = np.fft.fft(signal, axis=-1) / self._nbins / self._df
        self._qn = np.real(
            np.fft.ifft(self._v * self._phi) / self._norm * self._fs
        )

        _, self.t0_start, _ = self._OF.ofamp_withdelay()



    def update_signal(self, signal):
        """
        Function for updating all parameters that are signal-dependent,
        so that the pileup optimum filter can be run on new data
        without recalculating parameters unnecessarily.

        Parameters
        ----------
        signal : ndarray
            The signal that we want to apply the pileup optimum filter
            to (units should be Amps).

        Returns
        -------
        None

        """

        self._OF.update_signal(signal)
        self._update_signal(signal)


    def _determine_tcutoff(self, cutoff):
        """
        Hidden function to determine the cutoff on the time difference
        between pulses where, after this point, the iterative pileup
        optimum filter is 'good enough.'

        """

        tcutoff = np.argmax(self._pmatrix_off < cutoff) / self._fs

        return tcutoff


    def _get_amps(self, t0s):
        """
        Given the specified pulse times, this function returns the
        corresponding amplitudes, based on the OF formalism.

        """

        p_matrix = np.eye(len(t0s))
        ind_pairs = list(itertools.combinations(np.arange(len(t0s)), 2))

        if len(np.unique(t0s)) != len(t0s):
            pmatrix_inv = np.zeros((len(t0s), len(t0s)))
            pmatrix_inv[0, 0] = 1
        else:
            for ii, jj in ind_pairs:
                p_matrix[ii, jj] = p_matrix[jj, ii] = np.real(
                    np.dot(
                        self._phi,
                        self._s * np.exp(
                            -2.0j * np.pi * (t0s[jj] - t0s[ii]) * self._freqs
                        ),
                    ) / self._norm * self._df
                )
            pmatrix_inv = np.linalg.inv(p_matrix)

        qvec = np.zeros(len(t0s))

        for ii in range(len(t0s)):
            qvec[ii] = np.real(
                np.dot(
                    self._v,
                    self._phi * np.exp(
                        2.0j * np.pi * t0s[ii] * self._freqs
                    )
                ) / self._norm * self._df
            )

        return pmatrix_inv @ qvec

    def _chi2(self, t0s):
        """
        Given the specified pulse times, the chi-square is
        calculated.

        """

        numer = self._v - self._s * sum(
            a * np.exp(
                -2.0j * np.pi * t * self._freqs
            ) for a, t in zip(self._get_amps(t0s), t0s)
        )

        return np.real(np.dot(
            numer.conjugate() / self._psd,
            numer,
        ) * self._df)

    def _create_constraint(self, pulseconstraint):
        """
        Helper method for instantiating the constraint to be
        passed to `scipy.optimize.differential_evolution`. Also
        checks if the correct version of `scipy` is installed
        (1.4.0 or greater) to use this feature.

        """

        if pulseconstraint == 0:
            return dict()

        scipy_check = version.parse(SCIPY_VERSION) < version.parse('1.4.0')

        if pulseconstraint in [1, -1] and scipy_check:
            warnings.warn(
                'scipy must be version 1.4.0 or greater to use the '
                'pulseconstraint functionality. Your version is '
                f'{SCIPY_VERSION}. Defaulting to no pulseconstraint.'
            )
            return dict()

        if pulseconstraint == -1:
            constraints = optimize.NonlinearConstraint(
                lambda x: self._get_amps(x / self._fs),
                -np.inf,
                0,
            )
            return {'constraints' : constraints}

        if pulseconstraint == 1:
            constraints = optimize.NonlinearConstraint(
                lambda x: self._get_amps(x / self._fs),
                0,
                np.inf,
            )
            return {'constraints' : constraints}

        raise ValueError(
            'Unrecognized value passed to pulseconstraint, '
            'please pass 0, 1, or -1.'
        )


    def run(self, npulses, pulseconstraint=0, fit_window=None):
        """
        Runs the pileup optimum filter algorithm for the specified
        number of pulses.

        Parameters
        ----------
        npulses : int
            The total number of pulses to fit.
        pulseconstraint : int, optional
            Contrain the pulse direction of the pileup.
        fit_window : NoneType, tuple
            The indices in which to restrict the time shift window
            that the differential evolution algorithm will search
            for pulses in. Corresponds to range of time shift values
            (in unites of indices) to be allowed in the pulses, NOT 
            the time elapsed since the beginning of the event. If
            left as None, the value is determined by the estimate of
            where there is strong mixing of signals.

        Returns
        -------
        res : ndarray
            The results of the pileup optimum filter algorithm. The
            parameters returned are (amplitude 1, time offset 1,
            amplitude 2, time offset 2, ..., chi-square).

        """

        constraints = self._create_constraint(pulseconstraint)

        if fit_window is None:
            self._time_array = np.arange(
                -(int(npulses * self._tcutoff * self._fs)//2) + int(self.t0_start * self._fs),
                int(npulses * self._tcutoff * self._fs)//2 + int(
                    npulses * self._tcutoff * self._fs
                )%2 + int(self.t0_start * self._fs),
            ) / self._fs

            fit_window = (
                int(self._time_array[0] * self._fs),
                int(self._time_array[-1] * self._fs),
            )

        res = optimize.differential_evolution(
            lambda x: self._chi2(x / self._fs),
            npulses * (fit_window, ),
            **constraints,
        )

        t0s = np.sort(res['x'] / self._fs)
        amps = self._get_amps(t0s)
        chi2 = res['fun']
        self._res = res

        self.pileup_res = np.zeros(2 * npulses + 1)
        self.pileup_res[:-1:2] = amps
        self.pileup_res[1:-1:2] = t0s
        self.pileup_res[-1] = chi2

        return self.pileup_res



import numpy as np
from math import ceil, floor
from scipy.signal import savgol_filter
import matplotlib.pyplot as plt
from qetpy.utils import slope, fill_negatives, make_decreasing, fold_spectrum

__all__ = [
    "plot_psd",
    "plot_reim_psd",
    "plot_corrcoeff",
    "plot_csd",
    "plot_decorrelatednoise",
    "compare_noise",
    "plot_noise_sim",
]


def plot_psd(noise=None,
             psd=None, psd_freqs=None, channels=None,
             title=None,
             lgcoverlay=True, lgcsave=False, savepath=None,
             figsize=(8,5)):
    """
    Function to plot the noise spectrum referenced to the TES line in units of Amperes/sqrt(Hz).

    Parameters
    ----------
    noise : Object, optional
        Noise object to be plotted
    psd : 1D or 2D numpy array 
         psd in unit of Amps^2/Hz
    psd_freqs : 1D numpy array
        frequency array
    channels : list of string (optional)
        channel names, 
       If None, default names are "channel1", "channel2", etc
    title : str 
        title of the figure   
    lgcoverlay : boolean, optional
        If True, psd's for all channels are overlayed in a single plot,
        If False, each psd for each channel is plotted in a seperate subplot
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved
    figsize : tuple
        matplotlib figure size
    """

    # get correlation coefficient
    if noise is not None:
        psd = noise.psd
        if  psd is None:
            print('Need to calculate the corrcoeff first')
            return
        psd_freqs = noise.freqs
    elif psd is None or psd_freqs is None:
        raise ValueError('ERROR: correlation coeff matrix and '
                         'frequency array required!')

    # fold if needed
    has_negative_frequencies = np.any(psd_freqs < 0)
    if  has_negative_frequencies:
        max_positive_freq = np.max(psd_freqs)
        sample_rate = 2 * max_positive_freq
        psd_freqs, psd = fold_spectrum(psd, sample_rate)
    
    if psd.ndim == 1:
        psd = psd[np.newaxis, :]
    
    # channel names
    num_channels = psd.shape[0]
    if channels  is None:
        if noise is not None:
            channels = noise.channames
        else:
            channels = list()
            for ichan in range(num_channels):
                channels.append(f'Channel{ichan+1}')
                
    elif len(channels) != num_channels:
        raise ValueError(f'ERROR: number of channel names '
                         f'should be {num_channels}')

    if title is None:
        title = 'PSD'
        
    # Overlay plot
    if lgcoverlay:
        
        plt.figure(figsize=figsize)
        plt.title(title, fontweight='bold')
        plt.xlabel('Frequency [Hz]', fontweight='bold')
        plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]', fontweight='bold')
        plt.grid(which='both')
        for ichan, channel in enumerate(channels):
            plt.loglog(psd_freqs[1:], np.sqrt(psd[ichan][1:]),
                       label=channel)
            
        lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0)
        
        if lgcsave:
            try:
                plt.savefig(
                    savepath + '/PSD_overlay.png',
                    bbox_extra_artists=(lgd, ),
                    bbox_inches='tight',
                )
            except:
                print('WARNING: Invalid save path. Figure not saved')
        plt.show()

    else:
        num_subplots = len(channels)
        nrows = int(ceil(num_subplots / 2))
        ncolumns = 2
        figsize = (6*ncolumns, 4*nrows)            
        fig, axes = plt.subplots(nrows, ncolumns, figsize=figsize)
        plt.suptitle(title, fontweight='bold')
        for ii in range(nrows * 2):
            if ii < nrows:
                irow = ii
                jcolumn = 0
            else:
                irow = ii - nrows
                jcolumn = 1
            if ii < num_subplots and nrows > 1:    
                axes[irow,jcolumn].set_title(channels[ii], fontweight='bold')
                axes[irow,jcolumn].set_xlabel('Frequency [Hz]', fontweight='bold')
                axes[irow,jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]',
                                              fontweight='bold')
                axes[irow,jcolumn].grid(which='both')
                axes[irow,jcolumn].loglog(psd_freqs[1:], np.sqrt(psd[ii][1:]))
            elif ii < num_subplots and nrows==1:
                axes[jcolumn].set_title(channels[ii], fontweight='bold')
                axes[jcolumn].set_xlabel('Frequency [Hz]', fontweight='bold')
                axes[jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]',
                                         fontweight='bold')
                axes[jcolumn].grid(which = 'both')
                axes[jcolumn].loglog(psd_freqs[1:], np.sqrt(psd[ii][1:]))
            elif nrows==1:
                axes[jcolumn].axis('off')
            else:
                axes[irow,jcolumn].axis('off')
        plt.tight_layout() 
        plt.subplots_adjust(top=0.95)

        if lgcsave:
            try:
                plt.savefig(savepath + '/PSD_subplot.png')
            except:
                print('WARNINGL Invalid save path. Figure not saved')
        plt.show()



def plot_reim_psd(noise, lgcsave=False, savepath=None):
    """
    Function to plot the real vs imaginary noise spectrum referenced to the 
    TES line in units of Amperes/sqrt(Hz).
    This is done to check for thermal muon tails making it passed the quality cuts

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.real_psd is None:
        print('Need to calculate the psd first')
        return
    else:
        num_subplots = len(noise.channames)
        nrows = int(ceil(num_subplots / 2))
        ncolumns = 2
        fig, axes = plt.subplots(nrows, ncolumns, figsize=(6*num_subplots, 6*num_subplots)) 
        plt.suptitle('{} Real vs Imaginary PSD'.format(noise.name), fontsize=40)
        for ii in range(nrows * 2):
            if ii < nrows:
                irow = ii
                jcolumn = 0
            else:
                irow = ii - nrows
                jcolumn = 1
            if ii < num_subplots and nrows > 1:
                axes[irow,jcolumn].set_title(noise.channames[ii], fontsize=30)
                axes[irow,jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                axes[irow,jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]',
                                              fontsize=25)
                axes[irow,jcolumn].grid(which='both')
                axes[irow,jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.real_psd[ii][1:]), label='real')
                axes[irow,jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.imag_psd[ii][1:]), label='imag')
                axes[irow,jcolumn].legend()
            elif ii < num_subplots and nrows==1:
                axes[jcolumn].set_title(noise.channames[ii], fontsize=30)
                axes[jcolumn].set_xlabel('Frequency [Hz]', fontsize=25)
                axes[jcolumn].set_ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]',
                                         fontsize=25)
                axes[jcolumn].grid(which='both')
                axes[jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.real_psd[ii][1:]), label='real')
                axes[jcolumn].loglog(noise.freqs[1:], np.sqrt(noise.imag_psd[ii][1:]), label='imag')
                axes[jcolumn].legend()
            elif nrows==1:
                axes[jcolumn].axis('off')
            else:
                axes[irow,jcolumn].axis('off')
        plt.tight_layout() 
        plt.subplots_adjust(top=0.95)

        if lgcsave:
            try:
                plt.savefig(savepath + noise.name.replace(" ", "_") + '_ReIm_PSD.png')
            except:
                print('Invalid save path. Figure not saved')
        plt.show()        



def plot_corrcoeff(noise=None,
                   corrcoeff=None, corrcoeff_freqs=None,
                   channels=None,
                   lgcsmooth=True, nwindow=7, lgcsave=False,
                   savepath=None, figsize=(8,5)):
    """
    Function to plot the cross channel correlation coefficients. Since there are typically few traces,
    the correlations are often noisy. a savgol_filter is used to smooth out some of the noise

    Parameters
    ----------
    noise : QETpy Object (optional if "corrcoeff" not None)
        noise object to be plotted
    corrcoeff : 3D numpy array 
        correlation coefficient array [nchan, nchan, nfreqs]
    corrcoeff_freqs : 1D numpy array
        frequency array
    channels : list of string (optional)
        channel names, 
        if Nonem use "Channel1", "Channel2", etc
    lgcsmooth : boolean, optional
        If True, a savgol_filter will be used when plotting.
    nwindow : int, optional
        the number of bins used for the window in the savgol_filter
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    # get correlation coefficient
    if noise is not None:
        corrcoeff = noise.corrcoeff
        if corrcoeff is None:
            print('Need to calculate the corrcoeff first')
            return
        corrcoeff_freqs = noise.freqs
    elif corrcoeff is None or corrcoeff_freqs is None:
        raise ValueError('ERROR: correlation coeff matrix and '
                         'frequency array required!')

     
    # channel names
    num_channels = corrcoeff.shape[0]
    if channels  is None:

        if noise is not None:
            channels = noise.channames
        else:
            channels = list()
            for ichan in range(num_channels):
                channels.append(f'Channel{ichan+1}')
    elif len(channels) != num_channels:
        raise ValueError(f'ERROR: number of channel names '
                         f'should be {num_channels}')
    
    # figure
    plt.figure(figsize=figsize)
    plt.title(f'Cross Channel Correlation Coefficients', fontweight='bold')
    for ii in range(corrcoeff.shape[0]):
        for jj in range(corrcoeff.shape[1]):
            if ii > jj:
                label = f'{channels[ii]}-{channels[jj]}'
                if lgcsmooth:
                    plt.plot(
                        corrcoeff_freqs[1:],
                        savgol_filter(corrcoeff[ii][jj][1:], nwindow, 3,
                                      mode='nearest'),
                        label=label,
                        alpha=0.7,
                    )
                else:
                    plt.plot(corrcoeff_freqs[1:], corrcoeff[ii][jj][1:] ,
                             label=label, alpha=0.8)
                plt.xscale('log')
                    
    plt.xlabel('Frequency [Hz]', fontweight='bold')
    plt.ylabel(r'Correlation Coeff [COV(x,y)/$\sigma_x \sigma_y$]', fontweight='bold')
    plt.tick_params(which='both', direction='in', right=True, top=True)
    plt.grid(which='minor', linestyle='dotted')
    plt.grid(which='major')
    
    lgd = plt.legend(bbox_to_anchor=(1.03, 1), loc='best',
                     borderaxespad=1.)
        
    if lgcsave:
        try:
            plt.savefig(
                savepath + '/corrcoeff.png',
                bbox_extra_artists=(lgd, ),
                bbox_inches='tight',
            )
        except:
            print('Invalid save path. Figure not saved')

    plt.show()

def plot_csd(noise=None,
             csd=None, csd_freqs=None,
             channels=None, title=None,
             whichcsd=['02'], lgcreal=True,
             lgcsave=False, savepath=None,
             figsize=(8,5)):
    """
    Function to plot the cross channel noise spectrum referenced to the TES line in
    units of Amperes^2/Hz

    Parameters
    ----------
    noise : Object, optional
        Noise object to be plotted
    csd : 3D numpy array 
         csd in unit of Amps^2/Hz
    csd_freqs : 1D numpy array
        frequency array
    channels : list of string (optional)
        channel names, 
       If None, default names are "channel1", "channel2", etc
    title : str 
        title of the figure   
    whichcsd : list, optional
        a list of strings, where each element of the list refers to the pair of 
        indices of the desired csd plot
    lgcreal : boolean, optional
        If True, the Re(csd) is plotted. If False, the Im(csd) is plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """
    
    # get correlation coefficient
    if noise is not None:
        csd = noise.csd
        if csd is None:
            print('Need to calculate the csd first')
            return
        csd_freqs = noise.csd_freqs
    elif csd is None or csd_freqs is None:
        raise ValueError('ERROR: correlation coeff matrix and '
                         'frequency array required!')
     
    # channel names
    num_channels = csd.shape[0]
    if channels  is None:

        if noise is not None:
            channels = noise.channames
        else:
            channels = list()
            for ichan in range(num_channels):
                channels.append(f'Channel{ichan+1}')
    elif len(channels) != num_channels:
        raise ValueError(f'ERROR: number of channel names '
                         f'should be {num_channels}')


    # fold if needed
    has_negative_frequencies = np.any(csd_freqs < 0)
    if  has_negative_frequencies:
        max_positive_freq = np.max(csd_freqs)
        sample_rate = 2 * max_positive_freq
        csd_freqs, csd = fold_spectrum(csd, sample_rate)

    
    # get real/img csd
    real_csd = fill_negatives(np.real(csd))
    img_csd = fill_negatives(np.imag(csd))

    # display
    x_plt_label = []
    y_plt_label = []
    for label in whichcsd:
        if type(label) == str:
            x_plt_label.append(int(label[0]))
            y_plt_label.append(int(label[1]))
            if ((int(label[0]) > real_csd.shape[0] - 1)
                or (int(label[1]) > real_csd.shape[1] - 1)):
                print('index out of range')
                return
        else:
            print("Invalid selection. Please provide a list of strings "
                  "for the desired plots. Ex: ['01','02'] ")
            return

    for ii in range(len(x_plt_label)):
        plt.figure(figsize=figsize)
        if lgcreal:
            title = 'Re(CSD) for channels: {}-{}'.format(
                channels[x_plt_label[ii]],
                channels[y_plt_label[ii]],
            )

            plt.loglog(csd_freqs[1:], real_csd[x_plt_label[ii]][y_plt_label[ii]][1:])

        else:
            title = 'Im(CSD) for channels: {}-{}'.format(
                channels[x_plt_label[ii]],
                channels[y_plt_label[ii]],
            )

            plt.loglog(csd_freqs[1:], img_csd[x_plt_label[ii]][y_plt_label[ii]][1:])

        plt.title(title, fontweight='bold')
        plt.tick_params(which='both', direction='in', right=True, top=True)
        plt.grid(which='minor', linestyle='dotted')
        plt.grid(which='major')
        plt.xlabel('Frequency [Hz]', fontweight='bold')
        plt.ylabel(r'CSD [A$^2$/Hz]', fontweight='bold')

        if lgcsave:
            try:   
                plt.savefig(savepath + '/csd.png')
            except:
                print('Invalid save path. Figure not saved')
        plt.show()


def plot_decorrelatednoise(noise, lgcoverlay=False, lgcdata=True,
                           lgcuncorrnoise=True, lgccorrelated=False,
                           lgcsum=False, lgcsave=False, savepath=None,
                           figsize=(8,5)):
    """
    Function to plot the de-correlated noise spectrum referenced to the TES line in units of Amperes/sqrt(Hz) 
    from fitted parameters calculated calculate_deCorrelated_noise

    Parameters
    ----------
    noise : Object
        Noise object to be plotted
    lgcoverlay : boolean, optional
        If True, de-correlated for all channels are overlayed in a single plot,
        If False, the noise for each channel is plotted in a seperate subplot
    lgcdata : boolean, optional
        Only applies when lgcoverlay = False. If True, the csd data is plotted
    lgcuncorrnoise : boolean, optional
        Only applies when lgcoverlay = False. If True, the de-correlated noise is plotted
    lgccorrelated : boolean, optional
        Only applies when lgcoverlay = False. If True, the correlated component of the fitted noise
        is plotted
    lgcsum : boolean, optional
        Only applies when lgcoverlay = False. If True, the sum of the fitted de-correlated noise and
        and correlated noise is plotted
    lgcsave : boolean, optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """

    if noise.uncorrnoise is None:
        print('Need to de-correlate the noise first')
        return
    else:

        ### Overlay plot
        if lgcoverlay:
            plt.figure(figsize=figsize)
            plt.title('De-correlated noise', fontweight='bold')
            plt.xlabel('frequency [Hz]', fontweight='bold')
            plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]',
                       fontweight='bold')
            plt.tick_params(which='both', direction='in', right=True, top=True)
            plt.grid(which='minor', linestyle='dotted')
            plt.grid(which='major')
            for ichan, channel in enumerate(noise.channames):
                    plt.loglog(noise.freqs[1:], np.sqrt(noise.uncorrnoise[ichan][1:]),
                               label=channel)
            lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc='best',
                             borderaxespad=0)

            if lgcsave:
                try:
                    plt.savefig(
                        savepath + noise.name.replace(" ", "_") + '_deCorrNoise_overlay.png',
                        bbox_extra_artists=(lgd, ),
                        bbox_inches='tight',
                    ) 
                except:
                    print('Invalid save path. Figure not saved')
            plt.show()
        ### Subplots
        else:
            num_subplots = len(noise.channames)
            nrows = int(ceil(num_subplots / 2))
            ncolumns = 2
            figsize = (6*ncolumns, 4*nrows)            
            fig, axes = plt.subplots(nrows, ncolumns,
                                     figsize=figsize)

            plt.suptitle('De-correlated noise', fontweight='bold', y=1.03)
          
            alpha = 0.8
            if (lgcdata and lgcuncorrnoise and lgccorrelated):
                alpha = 0.5
                
            for ii in range(nrows * 2):
                if ii < nrows:
                    irow = ii
                    jcolumn = 0
                else:
                    irow = ii - nrows
                    jcolumn = 1
                if ii < num_subplots and nrows > 1:    
                    axes[irow,jcolumn].set_title(noise.channames[ii], fontweight='bold')
                    axes[irow,jcolumn].set_xlabel('Frequency [Hz]')
                    axes[irow,jcolumn].set_ylabel(r'Input Ref. Noise [A/$\sqrt{\mathrm{Hz}}$]')
                    axes[irow,jcolumn].tick_params(which='both', direction='in', right=True, top=True)
                    axes[irow,jcolumn].grid(which='minor', linestyle='dotted')
                    axes[irow,jcolumn].grid(which='major')
                    
                    if lgcdata:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.real_csd[ii][ii][1:]),
                            label='Data',
                            alpha=alpha,
                        )
                    if lgcuncorrnoise:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:]),
                            label='Uncorrelated Noise',alpha=alpha)
                    if lgccorrelated:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.corrnoise[ii][1:]),
                            label='Correlated Noise',
                            alpha=alpha,
                        )
                    if lgcsum:
                        axes[irow,jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:] + noise.corrnoise[ii][1:]),
                            label='Total Noise',
                            alpha=alpha,
                        )
                    axes[irow,jcolumn].legend()
                elif ii < num_subplots and nrows==1:
                    axes[jcolumn].set_title(noise.channames[ii], fontweight='bold')
                    axes[jcolumn].set_xlabel('Frequency [Hz]')
                    axes[jcolumn].set_ylabel(r'Input Ref. Noise [A/$\sqrt{\mathrm{Hz}}$]')
                    axes[jcolumn].grid(which='both')
                    if lgcdata:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.real_csd[ii][ii][1:]),
                            label='Data',
                            alpha=alpha,
                        )
                    if lgcuncorrnoise:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:]),
                            label='Uncorrelated Noise',
                            alpha=alpha,
                        )
                    if lgccorrelated:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.corrnoise[ii][1:]),
                            label='Correlated Noise',
                            alpha=alpha,
                        )
                    if lgcsum:
                        axes[jcolumn].loglog(
                            noise.freqs[1:],
                            np.sqrt(noise.uncorrnoise[ii][1:] + noise.corrnoise[ii][1:]),
                            label='Total Noise',
                            alpha=alpha,
                        )
                    axes[jcolumn].legend()
                elif nrows==1:
                    axes[jcolumn].axis('off')
                else:
                    axes[irow,jcolumn].axis('off')
            plt.tight_layout() 
            plt.subplots_adjust(top=0.95)
      
            if lgcsave:
                try:
                    plt.savefig(savepath + noise.name.replace(" ", "_") + '_deCorrNoise_subplot.png')
                except:
                    print('Invalid save path. Figure not saved')
                    
            plt.show()


def compare_noise(arr, channels, lgcdecorrelatednoise=False, lgcsave=False, savepath=None):
    """
    Function to plot multiple psds from different noise objects on the same figure. Each channel will
    be plotted in its own figure.

    Parameters
    ----------
    arr : array_like
        Array of noise objects
    channels : list
        List of strings, each string is a channel to plot. ex ['PSA1','PAS2']
    lgcdecorrelatednoise : boolean, optional
        If False, the psd is for each channel is plotted. If True, the calculated
        de-correlated noise is plotted
    lgcsave : boolean , optional
        If True, the figure is saved in the user provided directory
    savepath : str, optional
        Absolute path for the figure to be saved

    """


    #check to make sure channels is a list or array
    if (type(channels) == np.ndarray or type(channels) == list):
        for chan in channels:
            if chan in set(sum([arr[ii].channames for ii in range(len(arr))], [])): #check if channel is in 
                # any of the noise objects before creating a figure
                plt.figure(figsize=(12, 8))
                plt.title(chan) 
                plt.xlabel('Frequency [Hz]')
                plt.ylabel(r'Input Referenced Noise [A/$\sqrt{\mathrm{Hz}}$]')
                plt.grid(which = 'both')
                for ii in range(len(arr)):
                    if chan in arr[ii].chann_dict:
                        chan_index = arr[ii].chann_dict[chan] #check that the channel is in the noise object before plotting
                        #plot the de correlated noise if desired
                        if lgcdecorrelatednoise:
                            #check that de correlated noise has been calculated
                            if arr[ii].uncorrnoise is not None:
                                plt.loglog(
                                    arr[ii].freqs[1:],
                                    np.sqrt(arr[ii].uncorrnoise[chan_index][1:]),
                                    label=arr[ii].name + ' Decorrelated Noise',
                                )
                            else:
                                print('The decorrelated noise for file: {} has not been calculated yet'.format(arr[ii].name))
                        else:
                            plt.loglog(arr[ii].freqs[1:], np.sqrt(arr[ii].psd[chan_index][1:]), label=arr[ii].name)
                    else:
                        print('channel: {} not found for file: {} '.format(chan, arr[ii].name)) 
                lgd = plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0)
                if lgcsave:
                    try:
                        plt.savefig(
                            savepath + chan + '_PSD_comparison.png',
                            bbox_extra_artists=(lgd, ),
                            bbox_inches='tight',
                        )
                    except:
                        print('Invalid save path. Figure not saved')
                plt.show()
            else:
                print('Invalid channel name: {} '.format(chan))
    else:
        print("Please provide desired channels in format of a list of numpy.ndarray. ex ['PSA1','PAS2']")   





def plot_noise_sim(f, psd, noise_sim, istype, qetbias, lgcsave=False, figsavepath='', xlims=None, ylims=None):
    """
    Plots psd with simulated noise model

    Parameters
    ----------
    f : array_like
        Frequency bins for psd
    psd : array_like 
        Power spectral density
    istype : str
        Must be 'current', 'power', 'sc', or 'normal'
        If 'current' the noise is plotted referenced to TES current
        If 'power' the noise is plotted referenced to TES power
    qetbias : float
        Applied QET bias
    lgcsave : boolean, optional
        If True, plot is saved
    figsavepath : str, optional
        Directory to save trace
    xlims : NoneType, tuple, optional
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple, optional
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    if istype == 'current':
        fig, ax = _plot_ti_noise(f, psd, noise_sim, xlims, ylims)
        if lgcsave:
            fig.savefig(figsavepath+f'current_qetbias{qetbias*1e6:.3f}muA.png', bbox_inches='tight')
        else:
            return fig, ax 
    elif istype == 'power':
        fig, ax = _plot_tp_noise(f, psd, noise_sim, xlims, ylims)
        if lgcsave:
            fig.savefig(figsavepath+f'power_noise_qetbias{qetbias*1e6:.3f}muA.png', bbox_inches='tight')
        else:
            return fig, ax 
    elif istype == 'sc':
        fig, ax = _plot_sc_noise(f, psd, noise_sim, qetbias, xlims, ylims)    
        if lgcsave:
            plt.savefig(f'{figsavepath}SC_noise_qetbias{qetbias*1e6:.3f}muA.png')
        else:
            return fig, ax 
    elif istype == 'normal':
        fig, ax = _plot_n_noise(f, psd, noise_sim, qetbias, xlims, ylims)    
        if lgcsave:
            plt.savefig(f'{figsavepath}normal_noise_qetbias{qetbias*1e6:.3f}muA.png')
        else:
            return fig, ax 

def _plot_ti_noise(f, psd, noise_sim, xlims, ylims):
    """
    Helper function to plot transition noise in units of current

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    freqs = f[1:]
    psd = psd[1:]
    noise_sim.freqs = freqs

    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle='--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.tick_params(which="both", direction="in", right=True, top=True)
    ax.set_xlabel(r'Frequency [Hz]')
    ax.set_title(f"Current Noise For $R_0$ : {noise_sim.r0*1e3:.2f} $m\Omega$")

    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ites())),
        color='#1f77b4',
        linewidth=1.5,
        label='TES Johnson Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_iload())),
        color='#ff7f0e',
        linewidth=1.5,
        label='Load Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_itfn())),
        color='#2ca02c',
        linewidth=1.5,
        label='TFN Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_itot())),
        color='#d62728',
        linewidth=1.5,
        label='Total Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_isquid())),
        color='#9467bd',
        linewidth=1.5,
        label='Squid+Electronics Noise',
    )
    ax.loglog(freqs, np.sqrt(psd), color='#8c564b', alpha=0.8, label='Raw Data')

    ax.set_ylabel('TES Current Noise $[A/\sqrt{\mathrm{Hz}}]$')

    lgd = plt.legend(loc='upper right')

    return fig, ax
    
    
def _plot_tp_noise(f, psd, noise_sim, xlims, ylims):
    """
    Helper function to plot transition noise in units of power

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    freqs = f[1:]
    psd = psd[1:]
    noise_sim.freqs = freqs

    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle='--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.tick_params(which="both", direction="in", right=True, top=True)
    ax.set_xlabel(r'Frequency [Hz]')
    ax.set_title(f"Power Noise For $R_0$ : {noise_sim.r0*1e3:.2f} $m\Omega$")

    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptes())),
        color='#1f77b4',
        linewidth=1.5,
        label='TES Johnson Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_pload())),
        color='#ff7f0e',
        linewidth=1.5,
        label='Load Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptfn())),
        color='#2ca02c',
        linewidth=1.5,
        label='TFN Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_ptot())),
        color='#d62728',
        linewidth=1.5,
        label='Total Noise',
    )
    ax.loglog(
        noise_sim.freqs,
        np.sqrt(np.abs(noise_sim.s_psquid())),
        color='#9467bd',
        linewidth=1.5,
        label='Squid+Electronics Noise',
    )
    ax.loglog(
        freqs,
        np.sqrt(psd / (np.abs(noise_sim.dIdP(freqs))**2)),
        color='#8c564b',
        alpha=0.8,
        label='Raw Data',
    )
    ax.set_ylabel(r'Input Referenced Power Noise [W/$\sqrt{\mathrm{Hz}}$]')

    lgd = plt.legend(loc='upper right')

    return fig, ax

def _plot_sc_noise(f, psd, noise_sim, qetbias, xlims, ylims):
    """
    Helper function to plot SC noise 

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    f = f[1:]
    psd = psd[1:]
    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle = '--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.loglog(
        f, np.sqrt(psd), alpha=0.5, color='#8c564b', label='Raw Data',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_isquid(f)), color='#9467bd', label='Squid+Electronics Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_iloadsc(f)), color='#ff7f0e', label='Load Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itotsc(f)), color='#d62728', label='Total Noise',
    )
    ax.legend()
    ax.set_xlabel('Frequency [Hz]')
    ax.set_ylabel('Input Referenced Current Noise [A/$\sqrt{\mathrm{Hz}}$]')
    ax.set_title(f'Superconducting State noise for QETbias: {qetbias*1e6} $\mu$A')
    ax.tick_params(which="both", direction="in", right=True, top=True)

    return fig, ax    

def _plot_n_noise(f, psd, noise_sim, qetbias, xlims, ylims):
    """
    Helper function to plot normal state noise 

    Parameters
    ----------
    f : ndarray
        Array of frequency values
    psd : ndarray
        One sided Power spectral density
    noise_sim : TESnoise object
        The noise simulation object
    qetbias : float
        Applied QET bias
    xlims : NoneType, tuple
        Limits to be passed to ax.set_xlim()
    ylims : NoneType, tuple
        Limits to be passed to ax.set_ylim()    

    Returns
    -------
    fig : Object
        fig object from matplotlib.pyplot
    ax : Object
        ax object from matplotlib.pyplot

    """

    f = f[1:]
    psd = psd[1:]
    fig, ax = plt.subplots(figsize=(11, 6))
    if xlims is not None:
        ax.set_xlim(xlims)
    if ylims is not None:
        ax.set_ylim(ylims)

    ax.grid(which="major", linestyle = '--')
    ax.grid(which="minor", linestyle="dotted", alpha=0.5)
    ax.loglog(
        f, np.sqrt(psd), alpha=0.5, color='#8c564b', label='Raw Data',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_isquid(f)), color='#9467bd', label='Squid+Electronics Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itesnormal(f)), color='#1f77b4', label= 'TES Johnson Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_iloadnormal(f)), color='#ff7f0e', label='Load Noise',
    )
    ax.loglog(
        f, np.sqrt(noise_sim.s_itotnormal(f)), color='#d62728', label='Total Noise',
    )
    ax.legend()
    ax.set_xlabel('Frequency [Hz]')
    ax.set_ylabel('Input Referenced Current Noise [A/$\sqrt{\mathrm{Hz}}$]')
    ax.set_title(f'Normal State noise for QETbias: {qetbias*1e6} $\mu$A')
    ax.tick_params(which="both", direction="in", right=True, top=True)

    return fig, ax

